/* QNET to USB Adapter v1.9b, 20180208 for Teensy (++2.0)
 * Detect and capture network traffic on the 1-wire QNET port, forwarding to host PC via USB-SER interface
 * - 1.0: Message Management/RESET & STATUS
 * - 1.1: Basic SENDing capability
 * - 1.2: Basic RECeiVing capability
 * - 1.8: Extended SCOUT to 585us (from 550) when sending to better suit ZXNet/Spectrum which doesn't seem to catch our transmissions...
 * -TODO: FSERVE packet receive
 *      : Broadcast Send/Recv
*/

elapsedMicros timer, ledTimer;
// elapsedMillis ledTimer;

// Version constant
const char QNET_USB_VERSION[]="1.9b - 2018-02-08"; // NB - Remember to change this with each new version!

// Some timer constants

const byte debugTimingEntries=31;

const unsigned long MESSAGE_TIMEOUT=1500; // uS to wait for a new MQ Message (or at least the initial Signature)
const unsigned long MESSAGE_TIMEOUT_EXT=8000; // EXTended MQ Message timeout, when no other actions are pending (netPending/replyPending)

// PORT PIN definitions
const byte NET_IN    =0;  // D0 on Teensy ++2.0 (Brown)

const byte DEBUG_NET =1;  // D1 (Red)
const byte DEBUG_INT1=2;  // D2 (Yellow)
const byte DEBUG_INT2=3;  // D3 (Green)
const byte DEBUG_INT3=4;  // D4 (Blue)
const byte DEBUG_INT4=5;  // D5 (Purple)

const byte DEBUG_INT5=20; // B0 (White)

const byte NET_OUT   =7;  // D7

const byte LED_PIN_NET_RX=6; // LED pin (D3) - needs an additional LED fitted
const byte LED_PIN_NET_TX=6; // LED pin (D4) - needs an additional LED fitted
const byte LED_PIN_USB_RX=6; // LED pin (D5) - needs an additional LED fitted
const byte LED_PIN_USB_TX=6; // LED pin (D6) - built-in LED

const byte MAX_LED=4; // Count of LEDs in use
const unsigned int LED_FLASH_DURATION=63000; // 63 mS duty-period

// ZXNET SEND Bit timings (uS)
// - Works, though still room for improvement
const unsigned int BIT_TIME_SEND=15;           // 11uS per bit-cell
const unsigned int BIT0_TIME_SEND=20;          // 11 + 11/2 (ish) to centre of Bit-0

// QNET SEND Bit timings (uS)
//const unsigned int BIT_TIME_SEND=14;         // 11uS per bit-cell
//const unsigned int BIT0_TIME_SEND=20;        // 11 + 11/2 (ish) to centre of Bit-0

// RECV Bit timings (uS) - seem to be suited to both QNet and ZXNet
// - Original QNet timings were 11/18us
const unsigned int BIT_TIME_RECV=13;           // 11uS per bit-cell
const unsigned int BIT0_TIME_RECV=21;          // 11 + 11/2 (ish) to centre of Bit-0

const byte SCOUT_BITS=16;                 // Number of bits in SCOUT
const unsigned int SCOUT_BIT_TIME=43;     // 30uS per bit-cell during (TK2 compatible) SCOUT
const unsigned int SCOUT_BIT_TIME_DIV5=8; // 30/5 us while SCOUT_BIT being checked for contention
const unsigned int SCOUT_DETECT_TIME=50;  // 50us minimum continuous ACTive before SCOUT is recognised
const unsigned int SCOUT_SKIP_TIME=510;   // 500uS skip-time over a detected SCOUT (after the initial 50uS of ACTive)

const unsigned int SCOUT_WORD=0b1011111111111111; // 16x Bit SCOUT, with 2x START bits and 1x STOP (all ACTive)

const unsigned int LEAD_WAIT_TIME=2500;   // 2.5ms maximum wait for a H- or D- ACKnowledge or otherwise ACTive LEAD bits
const unsigned int LEAD_WAIT_TIME_2=500;  // 500us maximum wait for ACTive LEAD bits on subsequent bytes being read
const unsigned int START_WAIT_TIME=32000; // 32ms maximum wait INACTive START bit, following LEAD bit detection (Tk2 waits 260ms!)
const unsigned int PRE_ACK_DELAY=25;      // 25us Delay following receipt of HDR or DATA before sending H/DACK
//const unsigned int PRE_ACK_DELAY=10;      // 10us Delay following receipt of HDR or DATA before sending H/DACK (orig 25us turned in to 85us...)
const unsigned int POST_ACK_DELAY=125;    // 125us Delay following HACK before starting to send DATA packet
const byte VALID_ACK=0x01;    // ACKnowledge byte (currenlty only 0x01 is considered valid)

// QNET Bit lengths
const byte LEAD_BITS=2;  // Num of bits in LEADer of Byte-frame
const byte START_BITS=1; // Num of START bits of Byte-frame
const byte STOP_BITS=2;  // Num of STOP bits of Byte-frame
const byte FILL_BITS=1;  // Num of bits FILLing between consecutive Byte-framnes

// State Machine CONSTants
// const int PH_GAP=0;    // PHase 0: GAP
// const int PH_SCOUT=1;  // PHase 1: SCOUT
// const int PH_HEADER=2; // PHase 2: HEADER
// const int PH_HACK=3;   // PHase 3: Header ACKnowledge
// const int PH_DATA=4;   // PHase 4: DATA
// const int PH_DACK=5;   // PHase 5: Data ACKnowledge

// const int inactCont=0;  // Level/Edge: INACTive CONTinuous
// const int activeCont=1; // Level/Edge: ACTive CONTinuous
// const int activeEdge=2; // Level/Edge: ACTive EDGE
// const int inactEdge=3;  // Level/Edge: INACTive EDGE
// const int ignore=4;     // Level/Edge: IGNORE/skip

// Timings for each PHase: { [PH, Min Duration, Max Timeout] }
// const byte PH_TIMINGS_GAP[] = { inactCont,300,2000 };
// const byte PH_TIMINGS_SCOUT[] = { activeEdge,-1,3000, activeCont,50,-1, ignore,-1,480 };
// const byte PH_TIMINGS_HEADER[] = { inactEdge,-1,6000, inactCont,11,-1 };

const int MESSAGE_HLEN=10; // Length of each Message Header
const int MAX_MESSAGES=7; // Maximum number of pending Messages - 1
const int MAX_MESSAGE_ID=127; // Maximum MID before roll-over (back to 1)

const char MQ_SYNC[]= "%M$Q";
const int  MQ_SYNC_LEN=4;
const unsigned int MQ_SYNC_SUM=231; // Byte-additive int value of MQ_SYNC

// Offsets in to the MQ Message header (changed to all be ints - these are NOT the actual values, just offsets)
const int M_ID=0;
const int M_DEV=2;
const int M_ACT=3;
const int M_PARERR=4;
const int M_LENPAR2=6;
const int M_CHKSUM=8;

const byte M_DEV_BRIDGE=0; // DEVice ID for the Bridge
const byte M_DEV_RMS=1;    // DEVice ID for RMS
const byte M_DEV_NET=2;    // DEVice ID for NET
const byte M_DEV_MDV1=3;    // DEVice ID for MDV1_
const byte M_DEV_MDV2=4;    // DEVice ID for MDV2_

const byte M_ACT_RESET=0;      // Master Bridge RESET ACTion Code
const byte M_ACT_STATUS=1;     // Request Bridge STATUS ACTion Code
const byte M_ACT_RECEIVE=2;    // RECIEVE (NET) ACTion Code
const byte M_ACT_SEND=3;       // SEND (NET) ACTion Code
const byte M_ACT_SERV_START=6; // FSERVe START (NET) ACTion Code
const byte M_ACT_SERV_STOP=7 ; // FSERVe STOP (NET) ACTion Code
const byte M_ACT_UNKNOWN=255;  // Unknown ACTion code/CMD

const unsigned int MQ_REPLY_MASK = 0b1000000000000000; // Mask REPLY (bit 15) from M_ID (use &)
const unsigned int MQ_LONG_MASK  = 0b0100000000000000; // Mask LONG (bit 14) from M_ID (use &)

const unsigned int MAX_UINT = 65535;

const unsigned int MQ_ERR_NC     = 0b1111111111111111; // -1 : Not Complete - repeated NETwork error, cancelled
const unsigned int MQ_ERR_ORNG   = 0b1111111111111100; // -4 : Out of RaNGe - Unknown ACTion code/CMD
const unsigned int MQ_ERR_BFFL   = 0b1111111111111011; // -5 : BuFfer FuLl - No available space in SEND/RECV buffers
const unsigned int MQ_ERR_TRNS   = 0b1111111111110011; // -13: TRaNSmission error - Missing SYNC or Checksum error in USB data stream
const unsigned int MQ_ERR_IPAR   = 0b1111111111110001; // -15: Invalid PARameter - Invalid mAct or missing Message body/payload

const unsigned int MAX_PACKET_SZ=1020; // Maximum QNET packet size (255 x 4 for Server/Large packet)
const unsigned int REG_PACKET_SZ=255; // Reglar (max) packet size
const unsigned int PACKET_HDR_SZ=8;

const char *MQ_STATUS_TEXT_GEN[] = { "QNET-USB Bridge STATUS Reply", 
                                     "Firmware version",
                                     "Messages processed",
                                     "MQ Slots Total/Free",
                                     "Buffers available (Send/Recv)",
                                     "USB Bytes Sent/Rcvd (Drpd)",
                                     "NET packets Sent/Rcvd",
                                     "NET Bytes Sent/Rcvd",
                                     "FSERVE Listening state",
                                     "DEBUG" }; // STATUS Reply text

// As we are feeding NET_OUT through a PNP transitor to match the QLAN hardware,
//  we need to invert the usual meaning of ACTive and INACTive

const boolean NET_ACTive=LOW;
const boolean NET_INACTive=HIGH;

const byte RECV_OK=0;
const byte RECV_FAIL=1;
const byte RECV_UNEXPECTED=2;
const byte RECV_TIMEOUT=3;

const byte HDR_GOOD=0;
const byte HDR_CHKSUM_ERROR=1;
const byte HDR_NOT_FOR_ME=2;
const byte HDR_PHANTOM_BLOCK=3;

const byte SEND_OK=0;
const byte SEND_CONTENTION=1;
const byte SEND_NO_HACK=2;
const byte SEND_NO_DACK=3;

const byte ND_DEST=0;
const byte ND_SELF=1;
const byte ND_BLKL=2;
const byte ND_BLKH=3;
const byte ND_TYPE=4;
const byte ND_NBYT=5;
const byte ND_DCHK=6;
const byte ND_HCHK=7;

const byte NF_SBLK=254;

// Global Variables...

unsigned int debugTiming[debugTimingEntries];
byte nextDebugTimingEntry=0;

unsigned long ledStateArr[MAX_LED * 3] = { 3, 0, 0, 4, 0, 0, 5, 0, 0, 6, 0, 0 }; // LED State table 4x { LED_PIN, LED_STATE, LED_TIMER }

byte messageQueue[MAX_MESSAGES * MESSAGE_HLEN]; // Message Queue
unsigned int mQAvailableSlots;

char recvBuffer[MAX_PACKET_SZ + PACKET_HDR_SZ]; // Receive Buffer - from NET, to USB (x1 for now)
char sendBuffer[MAX_PACKET_SZ + PACKET_HDR_SZ]; // Send Buffer - from USB, to NET (x1 for now)

// int recvPendingQueue[MAX_MESSAGES]; // Recv Pending Queue {mQSlot }

byte myStationNumber, remoteStationNumber, cmd, netRecvPendCnt;
int nextFreeMQSlot, nextReply, nextNetSend, lastNetRcvd, fServeMQSlot;
boolean mQSlotAvailable, mQScanRequired;
boolean recvBufferFree, sendBufferFree;
boolean netActionPending, netSendPending, netRecvPending, replyPending;

boolean fServeListening;
char    fServeListeningState[6];

char debugStr[31];

unsigned int recvBuffersFree, sendBuffersFree;

unsigned long messagesProcessed, usbBytesRcvd, usbBytesDropped, usbBytesSent;
unsigned long netBytesRcvd, netBytesSent, netPacketsRcvd, netPacketsSent;

void setup() {
  // put your setup code here, to run once:
  
  Serial.begin(115200); // For Arduino, needs to match the host PC baud, for Teensy, effectively ignored
  pinMode(NET_IN, INPUT);
  pinMode(NET_OUT, OUTPUT);
  pinMode(DEBUG_NET,OUTPUT);
  pinMode(DEBUG_INT1, OUTPUT);
  pinMode(DEBUG_INT2, OUTPUT);
  pinMode(DEBUG_INT3, OUTPUT);
  pinMode(DEBUG_INT4, OUTPUT);
  pinMode(DEBUG_INT5, OUTPUT);
  pinMode(LED_PIN_NET_RX, OUTPUT);
  pinMode(LED_PIN_NET_TX, OUTPUT);
  pinMode(LED_PIN_USB_RX, OUTPUT);
  pinMode(LED_PIN_USB_TX, OUTPUT);
  
  digitalWrite(NET_OUT, NET_INACTive); // NB Actually HIGH to suit PNP transitor
  digitalWrite(DEBUG_NET, LOW);        // NET DEBUG port
  digitalWrite(DEBUG_INT1, LOW);       // INTernal Timing DEBUG port #1
  digitalWrite(DEBUG_INT2, LOW);       // INTernal Timing DEBUG port #2
  digitalWrite(DEBUG_INT3, LOW);       // INTernal Timing DEBUG port #3
  digitalWrite(DEBUG_INT4, LOW);       // INTernal Timing DEBUG port #4
  digitalWrite(DEBUG_INT5, LOW);       // INTernal Timing DEBUG port #5
  digitalWrite(LED_PIN_NET_RX, LOW);
  digitalWrite(LED_PIN_NET_TX, LOW);
  digitalWrite(LED_PIN_USB_RX, LOW);
  digitalWrite(LED_PIN_USB_TX, LOW);

}

void initialiseVariables() {
  
  // Initialise (clear) Message Queue (array)
  for (unsigned int messageIndex=0 ; messageIndex <= MAX_MESSAGES ; messageIndex++ ) {
    for (int byteIndex=0 ; byteIndex < MESSAGE_HLEN ; byteIndex++) {
      messageQueue[messageIndex * MESSAGE_HLEN + byteIndex]=0;
    }
  }

  // Initialise (clear) Receive and Send Buffers (arrays)
  for (unsigned int byteIndex=0 ; byteIndex < MAX_PACKET_SZ + PACKET_HDR_SZ ; byteIndex++ ) {
    recvBuffer[byteIndex]=0;
    sendBuffer[byteIndex]=0;
  } // for...

  // Initialise (clear) dynamic parts of ledStateArr { x, state, timer }
  for (unsigned int index=0 ; index < MAX_LED; index++ ) {
    ledStateArr[index * 3 + 1]=0;
    ledStateArr[index * 3 + 2]=0;
  } // for...
  
  replyPending=false;
  // digitalWrite(DEBUG_INT5, LOW); // Switch OFF DEBUG_INT5
  
  netSendPending=false;
  // digitalWrite(DEBUG_INT5, LOW); // Switch OFF DEBUG_INT5
  
  netRecvPending=false;
  netRecvPendCnt=0;
  netActionPending=false;
  
  nextReply=-1;
  nextNetSend=-1;

  nextFreeMQSlot=0;
  mQSlotAvailable=true;
  mQAvailableSlots=MAX_MESSAGES + 1;
  mQScanRequired=false;
  
  recvBufferFree=true;
  recvBuffersFree=1;
  sendBufferFree=true;
  sendBuffersFree=1;
  
  cmd=255;

  myStationNumber=1;
  remoteStationNumber=255;
  
  usbBytesRcvd=0;
  usbBytesDropped=0;
  usbBytesSent=0;
  messagesProcessed=0;
  fServeListening=false;
  fServeMQSlot=-1;
  strncpy(fServeListeningState, "FALSE", 5);
  
  strncpy(debugStr, "<none>", 6);
  
  netBytesRcvd=0;
  netBytesSent=0;
  netPacketsRcvd=0;
  netPacketsSent=0;
  
} // initialiseVariables()...

void flashLEDs(int iterations, int milSecs) {

  for (int flashes=0; flashes < iterations ; flashes++) {
    digitalWrite(LED_PIN_NET_RX, HIGH);
    digitalWrite(LED_PIN_NET_TX, HIGH);
    digitalWrite(LED_PIN_USB_RX, HIGH);
    digitalWrite(LED_PIN_USB_TX, HIGH);
    delay(milSecs);
    digitalWrite(LED_PIN_NET_RX, LOW);
    digitalWrite(LED_PIN_NET_TX, LOW);
    digitalWrite(LED_PIN_USB_RX, LOW);
    digitalWrite(LED_PIN_USB_TX, LOW);
    delay(milSecs);
  } // for...
  
} // flashLEDs()...

void flashLED_async() {

  byte thisLEDOffset;
  unsigned long thisLEDState, thisLEDPin, thisLEDTimer, thisLEDDuration;

  for (byte led=0; led < MAX_LED ; led++) {

    thisLEDOffset = led * 3;
    thisLEDState = ledStateArr[thisLEDOffset + 1];
    if ( thisLEDState == 0 ) continue; // Next LED, if not currently active
    
    thisLEDPin   = ledStateArr[thisLEDOffset];
    thisLEDTimer = ledStateArr[thisLEDOffset + 2];
    
    thisLEDDuration = ledTimer - thisLEDTimer;
    
    switch (thisLEDState) {

      case 1: // LED -ON- (already set by flashLED_start)

        if ( thisLEDDuration >= LED_FLASH_DURATION ) {
          digitalWrite(DEBUG_INT2, LOW);
          digitalWrite(thisLEDPin, LOW);
          ledStateArr[thisLEDOffset + 2] = ledTimer;
          ledStateArr[thisLEDOffset + 1]++;
        } // if...
        break;
        
      case 2: // LED -OFF-
        if ( thisLEDDuration >= LED_FLASH_DURATION ) {
          digitalWrite(DEBUG_INT2, LOW);
          digitalWrite(thisLEDPin, HIGH);
          ledStateArr[thisLEDOffset + 2] = ledTimer;
          ledStateArr[thisLEDOffset + 1]++;
        } // if...
        break;

      case 3: // LED -ON-

        if ( thisLEDDuration >= LED_FLASH_DURATION ) {
          digitalWrite(DEBUG_INT2, LOW);
          digitalWrite(thisLEDPin, LOW);
          ledStateArr[thisLEDOffset + 2] = ledTimer;
          ledStateArr[thisLEDOffset + 1]++;
        } // if...
        break;

      case 4: // LED -OFF-
        if ( thisLEDDuration >= LED_FLASH_DURATION ) {
          digitalWrite(DEBUG_INT2, LOW);
          digitalWrite(thisLEDPin, HIGH);
          ledStateArr[thisLEDOffset + 2] = ledTimer;
          ledStateArr[thisLEDOffset + 1]++;
        } // if...
        break;
        
      case 5: // LED -ON-
        if ( thisLEDDuration >= LED_FLASH_DURATION ) {
          digitalWrite(DEBUG_INT2, LOW);
          digitalWrite(thisLEDPin, LOW);
          ledStateArr[thisLEDOffset + 2] = ledTimer;
          ledStateArr[thisLEDOffset + 1]++;
        } // if...
        break;
        
      case 6: // LED -OFF-
        if ( thisLEDDuration >= LED_FLASH_DURATION ) {
          digitalWrite(DEBUG_INT2, LOW);
          digitalWrite(thisLEDPin, HIGH);
          ledStateArr[thisLEDOffset + 2] = ledTimer;
          ledStateArr[thisLEDOffset + 1]++;
        } // if...
        break;

      case 7: // LED -ON-
        if ( thisLEDDuration >= LED_FLASH_DURATION ) {
          digitalWrite(DEBUG_INT2, LOW);
          digitalWrite(thisLEDPin, LOW);
          ledStateArr[thisLEDOffset + 2] = 0;
          ledStateArr[thisLEDOffset + 1] = 0; // Reset State to zero, with LED -OFF-
        } // if...
        break;
        
    } // switch...

  } // for...
  
  digitalWrite(DEBUG_INT2, HIGH);
  
} // flashLED_async()...

void flashLED_start(byte ledPin) {

  boolean ledPinFound=false;
  byte led, thisLEDOffset;

  for ( led=0; led < MAX_LED; led++ ) {
  
    if ( ledStateArr[led * 3] == long(ledPin) ) {
      
      ledPinFound=true;
      thisLEDOffset = led * 3;
      break;
      
    } // if...
  
  } // for...

  if ( !ledPinFound ) return;

  digitalWrite(ledPin, HIGH);
  ledStateArr[thisLEDOffset + 1] = 1;
  ledStateArr[thisLEDOffset + 2] = ledTimer;
  
} // flashLED_start()...

void putMQ(int mQSlot, unsigned int mID, byte mDev, byte mAct, unsigned int mParErr, unsigned int mLenPar2, unsigned int mChkSum) {
  // Update Message Queue at mQSlot

  int mQOffset;

  if ( (mQSlot > MAX_MESSAGES) || (mQSlot < 0) ) return; // Out of bounds
  
  mQOffset = mQSlot * MESSAGE_HLEN;
  messageQueue[mQOffset]                 = highByte(mID);
  messageQueue[mQOffset + 1]             = lowByte(mID);
  messageQueue[mQOffset + M_DEV]         = mDev;
  messageQueue[mQOffset + M_ACT]         = mAct;
  messageQueue[mQOffset + M_PARERR]      = highByte(mParErr);
  messageQueue[mQOffset + M_PARERR + 1]  = lowByte(mParErr);
  messageQueue[mQOffset + M_LENPAR2]     = highByte(mLenPar2);
  messageQueue[mQOffset + M_LENPAR2 + 1] = lowByte(mLenPar2);
  messageQueue[mQOffset + M_CHKSUM]      = highByte(mChkSum);
  messageQueue[mQOffset + M_CHKSUM + 1]  = lowByte(mChkSum);
  
} // putMQ...

void markMQSlotEmpty(int mQSlot) {
  // Update Message Queue at mQSlot with mID=0;

  int mQOffset;

  if ( (mQSlot > MAX_MESSAGES) || (mQSlot < 0) ) return; // Out of bounds
  
  mQOffset = mQSlot * MESSAGE_HLEN;
  
  for ( byte byteIndex=0; byteIndex < MESSAGE_HLEN; byteIndex++ ) messageQueue[mQOffset + byteIndex] = 0;
  
} // markMQSlotEmpty()...

void getMQ(int mQSlot, unsigned int &mID, byte &mDev, byte &mAct, unsigned int &mParErr, unsigned int &mLenPar2, unsigned int &mChkSum) {
  /* Read-back Message Queue details from mQSlot
   * NB - Use of simple dereferencing '&' to allow passing-back values without resorting to pointers.
  */
  
  int mQOffset;

  if ( (mQSlot > MAX_MESSAGES) || (mQSlot < 0) ) return; // Out of bounds
  
  mQOffset = mQSlot * MESSAGE_HLEN;

  mID      = word( messageQueue[mQOffset], messageQueue[mQOffset + 1] );
  mDev     = messageQueue[mQOffset + M_DEV];
  mAct     = messageQueue[mQOffset + M_ACT];
  mParErr  = word( messageQueue[mQOffset + M_PARERR],  messageQueue[mQOffset + M_PARERR + 1] );
  mLenPar2 = word( messageQueue[mQOffset + M_LENPAR2], messageQueue[mQOffset + M_LENPAR2 + 1] );
  mChkSum  = word( messageQueue[mQOffset + M_CHKSUM],  messageQueue[mQOffset + M_CHKSUM + 1] );
  
} // getMQ...

unsigned int calculateMQChkSum(unsigned int mID, byte mDev, byte mAct, unsigned int mParErr, unsigned int mLenPar2) {
  // Calculate the byte-additive ChkSum of MQ_SYNC + mID + mDev + mAct + mParErr + mLenPar2

  unsigned int chkSum = MQ_SYNC_SUM;
  
  chkSum+= highByte(mID);
  chkSum+= lowByte(mID);
  chkSum+= mDev;
  chkSum+= mAct;
  chkSum+= highByte(mParErr);
  chkSum+= lowByte(mParErr);
  chkSum+= highByte(mLenPar2);
  chkSum+= lowByte(mLenPar2);

  return chkSum;
  
} // calculateMQChkSum()...

void queueMessage() {
  // Check USB for a new Message and queue it in nextFreeMQSlot if available (mQSlotAvailable=true)

  boolean messageReceived, mQSyncFound, longMessageExpected, longMessageRcvd;
  boolean bufferError;

  int nextMQSyncIndex, mQOffset;

  byte mDev, mAct, nextMQSyncByte, nextByte;

  unsigned int mQTimeout, mID, mParErr, mLenPar2, mChkSum, newChkSum;
  unsigned int byteIndex;
  
  if ( !mQSlotAvailable ) return;

  mQScanRequired=true;
  
  messageReceived=false;
  longMessageExpected=false;
  longMessageRcvd=false;
  bufferError=false;
  
  if ( netActionPending || replyPending ) {
  
    mQTimeout = MESSAGE_TIMEOUT; // 1.5ms
    
  } else {
  
    mQTimeout = MESSAGE_TIMEOUT_EXT;  // 8ms
    
  }

  timer=0; // Reset the timer - we only stay in here until a new message is received, or else until mQTimeout expires

  while ( timer < mQTimeout && !messageReceived ) {
    //  Terminate reading USB if timeout without a valid MQSync/Header being found, or else when a full message is read

    mQSyncFound=false;
    nextMQSyncIndex=0;

    // First, scan for MQ Sync Signature
    do {
    
      if ( !Serial.available() ) continue;
      
      nextMQSyncByte=byte( MQ_SYNC[nextMQSyncIndex] );

      nextByte=Serial.read();
      usbBytesRcvd++;
      
      if ( nextByte != nextMQSyncByte ) {
      
        // flashLEDs(2, 1000);
        usbBytesDropped += nextMQSyncIndex;

  digitalWrite(DEBUG_INT5, HIGH); // Switch ON DEBUG_INT5
  delayMicroseconds(10);
  digitalWrite(DEBUG_INT5, LOW); // Switch OFF DEBUG_INT5
        
        if ( nextByte == byte(MQ_SYNC[0]) ) {
        
          nextMQSyncIndex=1;
          continue; // Accomodate a repeated match at the beginning of MQ_SYNC, without losing the most recent captured byte from the data-stream - takes a bit of thinking about...
        
        } else {   

          usbBytesDropped++; // adjust the number of bytes dropped...   
          nextMQSyncIndex=0; // Start over from beginning of MQ_SYNC, from next byte in the data-stream          

          continue;
          
        } // if...
        
      } // if...
      
      if ( nextMQSyncIndex < (MQ_SYNC_LEN - 1) ) {
      
        nextMQSyncIndex++;
        
      } else {
      
        mQSyncFound=true;
        
      } // if...
      
    } while ( timer < mQTimeout && !mQSyncFound ); // do...while...

    if ( !mQSyncFound ) {
    
      return; // We timed-out looking for a fresh MQ Sync Signature
      
    }

    /* So, we've found the MQ Sync signature and we haven't yet timed-out - read MQ Header with an extended timeout...
    *   NB - it is still quite possible that the byte stream we found was NOT a real SYNC and, whilst the ChkSum check later should cause this block
    *        to be discarded, it is less likely, but possible, that the 'real' SYNC is contained in what we capture as the Header - and will thus be missed.
    *        Unless, of course, we go to some lengths to scan the header for a potential SYNC as well. Thanks, but no thanks...
    */
    
    mQOffset = nextFreeMQSlot * MESSAGE_HLEN;
    byteIndex=0; // replacement for for-loop, so we can 'continue' if no bytes yet available from USB without incrementing byteIndex

    digitalWrite(DEBUG_INT1, LOW); // Switch OFF DEBUG_INT1
    delayMicroseconds(10);
    digitalWrite(DEBUG_INT1, HIGH); // Switch ON DEBUG_INT1

    do {

      if ( timer > MESSAGE_TIMEOUT_EXT ) {

  // mQScanRequired=false; // Nothing new in the MQ to talk-about...
  
  if ( byteIndex > 0 ) {
  
    markMQSlotEmpty( nextFreeMQSlot ); // Mark (partially filled) MQ Slot 'empty'
          
        }
        
        usbBytesDropped += ( MQ_SYNC_LEN + byteIndex );
        
  digitalWrite(DEBUG_INT5, HIGH); // Switch ON DEBUG_INT5
  delayMicroseconds(20);
  digitalWrite(DEBUG_INT5, LOW); // Switch OFF DEBUG_INT5
        
  return; // We timed-out waiting for a (complete) Message Header

      } // if...

      if ( !Serial.available() ) continue;

      messageQueue[mQOffset + byteIndex] = Serial.read();
      usbBytesRcvd++;
      byteIndex++;

    } while ( byteIndex < MESSAGE_HLEN ); // do...while...
    
    digitalWrite(DEBUG_INT1, LOW); // Switch OFF DEBUG_INT1
    delayMicroseconds(20);
    digitalWrite(DEBUG_INT1, HIGH); // Switch ON DEBUG_INT1

    // OK, got the header of a new message in to messageQueue at nextFreeMQSlot - lets first check ChkSum and, if there's more to read, do so

    getMQ( nextFreeMQSlot, mID, mDev, mAct, mParErr, mLenPar2, mChkSum );

    newChkSum = calculateMQChkSum( mID, mDev, mAct, mParErr, mLenPar2 );

    if ( mChkSum != newChkSum ) {

      // flashLEDs(4, 1000);
      usbBytesDropped += ( MQ_SYNC_LEN + MESSAGE_HLEN );
      markMQSlotEmpty( nextFreeMQSlot ); // Mark MQ Slot 'empty'
      
      digitalWrite(DEBUG_INT5, HIGH); // Switch ON DEBUG_INT5
      delayMicroseconds(30);
      digitalWrite(DEBUG_INT5, LOW); // Switch OFF DEBUG_INT5

      continue; // Check sum didn't match sync+data, continue to scan data stream until we find another MQ Sync, or timeout (what else can we do?!)

    } // if...
    
    digitalWrite(DEBUG_INT1, LOW); // Switch OFF DEBUG_INT1
    delayMicroseconds(30);
    digitalWrite(DEBUG_INT1, HIGH); // Switch ON DEBUG_INT1

    if ( ( ( mDev == M_DEV_NET && 
         ( ( mAct == M_ACT_RECEIVE ) ||
     ( mAct == M_ACT_SEND ) ||
     ( mAct == M_ACT_SERV_START ) || 
     ( mAct == M_ACT_SERV_STOP ) ) )
       ||
       ( mDev == M_DEV_BRIDGE && 
         ( ( mAct == M_ACT_RESET ) || 
     ( mAct == M_ACT_STATUS ) ) ) 
       ) ) {

      cmd=mAct;
      if (cmd == M_ACT_SEND ) {

  longMessageExpected=true;
  if ( !sendBufferFree || (mLenPar2 > (MAX_PACKET_SZ + PACKET_HDR_SZ)) ) bufferError=true;

      } // if...

    } else {

      cmd=M_ACT_UNKNOWN; // Only dealing with NET:RECEIVE, SEND, SERV_START, SERV_STOP & BRIDGE:RESET, STATUS device requests today...

    } // if...

    messageReceived=true; // MQ Header looks good - read rest of message-body (if any) assuming sendBufferFree
    messagesProcessed++;

    if ( mID & MQ_LONG_MASK ) {

      // LONG Message - try to capture the Message body in to next available sendBuffer if required, or else purge
      longMessageRcvd=true;

      if ( longMessageExpected && !bufferError ) {
  // Capture to sendBuffer

  for ( byteIndex=0; byteIndex < mLenPar2; byteIndex++ ) {

    if ( Serial.available() ) {

      sendBuffer[byteIndex] = Serial.read();
      usbBytesRcvd++;

    } else {

      longMessageRcvd=false;
      break;

    } // if...

  } // for...

  if ( longMessageRcvd ) {

    sendBuffersFree--;
    if ( sendBuffersFree < 1 ) sendBufferFree=false;
    
    digitalWrite(DEBUG_INT1, LOW); // Switch OFF DEBUG_INT1
    delayMicroseconds(40);
    digitalWrite(DEBUG_INT1, HIGH); // Switch ON DEBUG_INT1

  } // if...

      } else {
  // Can't - or don't need to - store the message body - just purge what's avaialble

  longMessageRcvd=false;

  for ( byteIndex=0; byteIndex < mLenPar2; byteIndex++ ) {

    if ( Serial.available() ) {

      nextByte=Serial.read();
      usbBytesDropped++;

    } else {

      break;

    } // if...           

  } // for...

      } // if...

    } // if...
      
    if ( messageReceived ) {
    
      flashLED_start(LED_PIN_USB_RX);
    
      mQAvailableSlots--;
      if ( mQAvailableSlots < 1  ) mQSlotAvailable=false;

      netActionPending=false;
      replyPending=false;
      // digitalWrite(DEBUG_INT5, LOW); // Switch OFF DEBUG_INT5

      switch (cmd) {

        case M_ACT_RESET:
          // Mark the Message for REPLY and prioritise for next reply action (no further replies will then take place - due to re-initialisation)
          
          mID = mID | MQ_REPLY_MASK; // Set top bit to mark as REPLY
          mID = mID & ~MQ_LONG_MASK; // Clear bit-14 to mark as SHORT
          mParErr=0; // No ERRor/OK
          mLenPar2=0;
          newChkSum = calculateMQChkSum( mID, mDev, mAct, mParErr, mLenPar2 );
          
          putMQ( nextFreeMQSlot, mID, mDev, mAct, mParErr, mLenPar2, newChkSum );
          
          nextReply = nextFreeMQSlot;
          netActionPending=false;
          netSendPending=false;
          netRecvPending=false;
          replyPending=true;
          // digitalWrite(DEBUG_INT5, HIGH); // Switch ON DEBUG_INT5
          mQScanRequired=false;
          
          return; // No need to revisit the MQ table as we'll re-initialise immediately once the REPLY is sent...
 
        case M_ACT_STATUS:
          // Add a STATUS REPLY Message to the MQ
          
          if ( !recvBufferFree ) {
          
            mID = mID | MQ_REPLY_MASK; // Set bit-15 to mark as REPLY
            mID = mID & ~MQ_LONG_MASK; // Clear bit-14 to mark as SHORT
            mParErr = MQ_ERR_BFFL; // ERRor Buffer Full
            mLenPar2=0;

          } else {
          
            mID = mID | MQ_REPLY_MASK; // Set bit-15 to mark as REPLY
            mID = mID | MQ_LONG_MASK;  // Set bit-14 to mark as LONG
            mParErr=0; // No ERRor/OK
            
            /* Queue the STATUS REPLY here...
            *    QNET-USB Bridge STATUS Reply
            *    ----------------------------
            *    Firmware version:
            *    Messages processed:
            *    MQ Slots Total/Free:
            *    Buffers available Send/Recv:
            *    USB Bytes Sent/Rcvd (Drpd):
            *    NET packets Sent/Rcvd:
            *    NET Bytes Sent/Rcvd:
            *    FSERVE Listening state:
            *    DEBUG:
            */
            
            mLenPar2 = snprintf( recvBuffer, MAX_PACKET_SZ + PACKET_HDR_SZ - 1,
                       "%s\n----------------------------\n%s: %s\n%s: %lu\n%s: %u/%u\n%s: %u/%u\n%s: %lu/%lu (%lu)\n%s: %lu/%lu\n%s: %lu/%lu\n%s: %s\n%s: %s",
                       MQ_STATUS_TEXT_GEN[0],
                       MQ_STATUS_TEXT_GEN[1], QNET_USB_VERSION,
                       MQ_STATUS_TEXT_GEN[2], messagesProcessed,
                       MQ_STATUS_TEXT_GEN[3], MAX_MESSAGES + 1, mQAvailableSlots,
                       MQ_STATUS_TEXT_GEN[4], sendBuffersFree, recvBuffersFree,
                       MQ_STATUS_TEXT_GEN[5], usbBytesSent, usbBytesRcvd, usbBytesDropped,
                       MQ_STATUS_TEXT_GEN[6], netPacketsSent, netPacketsRcvd,
                       MQ_STATUS_TEXT_GEN[7], netBytesSent, netBytesRcvd,
                       MQ_STATUS_TEXT_GEN[8], fServeListeningState,
                       MQ_STATUS_TEXT_GEN[9], debugStr
                       );

            recvBuffersFree--;
            if ( recvBuffersFree < 1 ) recvBufferFree=false;

          } // if...
          
          newChkSum = calculateMQChkSum( mID, mDev, mAct, mParErr, mLenPar2 );
          putMQ( nextFreeMQSlot, mID, mDev, mAct, mParErr, mLenPar2, newChkSum );

          nextReply = nextFreeMQSlot;
          netSendPending=false;
          netRecvPending=false;
          netActionPending=false;
          replyPending=true;
          // digitalWrite(DEBUG_INT5, HIGH); // Switch ON DEBUG_INT5
          mQScanRequired=false;

          return; // No need to revisit the MQ table - we'll prioritise the STATUS reply
        
        case M_ACT_RECEIVE:
          /* DST.b & SRC.b in mParErr, Block.l & Block.h in mLenPar2.w
           *  - nothng to do here for now...
          */

          break;
        
        case M_ACT_SEND:
          /* Expect the Packet to SEND in the message body - if available, it has been read in to sendBuffer
           *  - otherwise REPLY with error... ERR_IPAR (-15) or ERR_BFFL (-13)
          */
          
          if ( !longMessageRcvd ) {
            // We expect LONG message - if not, it's a protocol error - send an ERRor REPLY
            
            mID = mID | MQ_REPLY_MASK;  // Set bit-15 to mark as REPLY
            mID = mID & ~MQ_LONG_MASK;  // Clear bit-14 to mark SHORT
            
            if ( bufferError ) {
            
              mParErr = MQ_ERR_BFFL; // ERRor BuFfer FuLl
              
            } else {
            
              mParErr = MQ_ERR_IPAR; // ERRor Invalid PARameter
              
            } // if...

            mLenPar2=0;
            newChkSum = calculateMQChkSum( mID, mDev, mAct, mParErr, mLenPar2 );
            putMQ( nextFreeMQSlot, mID, mDev, mAct, mParErr, mLenPar2, newChkSum );
            
          } // if...

          break;
        
        case M_ACT_SERV_START:
          // Add a persistent FSERVE Message to the MQ - like a RECEIVE, but won't be removed after REPLYing with the received packet - no error if already set
          
          if ( fServeListening ) {
            // Already listening for FSERVE packets - mark this message 'empty'
            
            markMQSlotEmpty( nextFreeMQSlot );
            
          } else {
          
            fServeListening=true;
            strncpy(fServeListeningState, "TRUE ", 5);
            fServeMQSlot=nextFreeMQSlot;
            
          }

          break;
        
        case M_ACT_SERV_STOP:
          // Remove persistent FSERVE Message from the MQ, if it exists (no error if not)
          
          markMQSlotEmpty( nextFreeMQSlot );
          
          if ( fServeListening ) {
            // Was listening for FSERVE packets - mark that message (fServeMQSlot) also as 'empty'
            
            markMQSlotEmpty( fServeMQSlot );
            fServeListening=false;
            strncpy(fServeListeningState, "FALSE", 5);
            fServeMQSlot=-1;
            
          }          

          break;

        case M_ACT_UNKNOWN:
        
          flashLEDs(3, 2000);
          
        default:
          // Unknown CMD in Message - prepare REPLY with Unknown CMD ERRor (ORNG)
          
          mID=mID | MQ_REPLY_MASK;
          mID=mID & ~MQ_LONG_MASK;
          mParErr=MQ_ERR_ORNG; // ERRor Out of RaNGe
          mLenPar2=0;

          newChkSum = calculateMQChkSum( mID, mDev, mAct, mParErr, mLenPar2 );
          
          putMQ( nextFreeMQSlot, mID, mDev, mAct, mParErr, mLenPar2, newChkSum );
          
          break;
          
      } // switch...
      
    } // if...
       
  } // while...
  
} // queueMessage()...

void scanMessageQueue() {
  // Now find next available MQ Slot (M_ID's == 0), any pending Replies or netActions...

  boolean thisReplyPending, thisNetSendPending;
  byte mDev, mAct;
  unsigned int mID, mParErr, mLenPar2, mChkSum, lowestReplySoFar, lowestNetSendSoFar;
  unsigned int mIDMasked;
  
  mQSlotAvailable=false;
  mQAvailableSlots=0;
  
  netActionPending=false;
  
  netSendPending=false;
  // digitalWrite(DEBUG_INT5, LOW); // Switch OFF DEBUG_INT5
  
  netRecvPending=false;
  
  nextNetSend=-1;
  netRecvPendCnt=0;
  
  lowestReplySoFar=MAX_UINT;
  lowestNetSendSoFar=MAX_UINT;
    
  for ( int mQSlot=0; mQSlot <= MAX_MESSAGES; mQSlot++ ) {

    getMQ( mQSlot, mID, mDev, mAct, mParErr, mLenPar2, mChkSum );
    
    mIDMasked = mID       & ~MQ_REPLY_MASK;
    mIDMasked = mIDMasked & ~MQ_LONG_MASK;
    
    if ( mIDMasked == 0 ) {
      // This SHOULDN'T happen...
      
      markMQSlotEmpty( mQSlot );
      mID = 0;
      
    } // if...
    
    if ( mID == 0 ) {
    
      mQAvailableSlots++;
      
      if ( !mQSlotAvailable ) {
      
        nextFreeMQSlot = mQSlot; // Find the lowest free MQ Slot
        mQSlotAvailable = true;
        
      } // if...
      
      continue; // This slot is 'empty' - next mQSlot...
      
    } // if...

    thisReplyPending=false;
    thisNetSendPending=false;

    if ( mID & MQ_REPLY_MASK ) thisReplyPending = true;
    
    if ( ( mAct == M_ACT_RECEIVE || mAct == M_ACT_SERV_START ) && !thisReplyPending ) {

      netRecvPendCnt++;
      netRecvPending = true;
      
    } // if
    
    if ( (mAct == M_ACT_SEND) && !thisReplyPending ) thisNetSendPending = true;
    
    mID = mIDMasked; // Mask-out the REPLY and LONG bits (temporarily - still marked as such in the MQ record)
    
    if ( thisReplyPending && (mID < lowestReplySoFar) ) {

      replyPending = true;
      // digitalWrite(DEBUG_INT5, HIGH); // Switch ON DEBUG_INT5
      nextReply = mQSlot;
      lowestReplySoFar = mID;
      continue; // Next mQSlot...

    } // if...

    if ( thisNetSendPending && (mID < lowestNetSendSoFar) ) {
      
      netSendPending = true;
      // digitalWrite(DEBUG_INT5, HIGH); // Switch ON DEBUG_INT5
      nextNetSend = mQSlot;
      lowestNetSendSoFar = mID;
            
    } // if...

  } // for...
  
  if ( netSendPending || netRecvPending ) netActionPending=true;
  
} // scanMessageQueue...

void actionNET() {
  /* Depending upon netSend and netRecvPending flags, listen/transmit across the wire:
   *  netSend: check for 3ms of continuous inactivity, then transmit nextNetSend message
   *  netRecv: check for at least 200us and no more than 20ms of inactivity
   *  netSend & netRecv: check for at least 200us of inactivity and not more than 3ms, then switch to transmit instead
  */

  boolean gapFound, packetReceived, packetSent;

  unsigned int minRecvGap, maxRecvGap, minSendGap;
  unsigned int mID, mParErr, mLenPar2, mChkSum, newChkSum;
  
  byte packetRecvCode, packetSendCode;
  byte mDev, mAct;
  
  minRecvGap=200;   // Minimum GAP before we'll recognise a packet for receive
  maxRecvGap=20000; // Maximum GAP before we timeout waiting for a packet
  minSendGap=3000;  // Minimum GAP before we'll start to transmit a packet

  gapFound=false;
  packetReceived=false;
  lastNetRcvd=-1;
  packetSent=false;
  
  packetRecvCode=-1;
  packetSendCode=-1;

  if ( netSendPending && netRecvPending ) {
    
    gapFound = waitGap(minRecvGap);
    if ( gapFound == false ) return; // no minimum GAP found - just quit for now (retrying on next cycle)
      
    packetRecvCode=listenForPacket(minSendGap - minRecvGap, lastNetRcvd);
    // NB lastNetRcvd is passed by &reference

    switch (packetRecvCode) {

      case RECV_OK:
        // Packet received OK (no chksum errors) and (started) within the time-out
        // lastNetRcvd will have been updated with the MQ Slot ID needed for the reply

        netPacketsRcvd++;
        packetReceived=true;
        break;

      case RECV_FAIL:
        // Packet detected but either chksum mismatch or some other framing error

        return;

      case RECV_UNEXPECTED:
        // Packet received successfully, but not one that we were expecting (Dst/Src & Blk-L/Blk-H)
        
        return;
      
      case RECV_TIMEOUT:
        // No packet received, let's move on to transmit instead (minSendGap reached...)

        packetSendCode=sendPacket(nextNetSend);
        switch (packetSendCode) {

          case SEND_OK:
            // Packet successfully sent/acknowledged

            netPacketsSent++;
            sendBuffersFree++;
            sendBufferFree=true;
            packetSent=true;
            break;

          case SEND_CONTENTION:
            // Send failed due to contention detected during SCOUT, quit for now (retrying on next cycle)

            return;

          case SEND_NO_HACK:
            // Send failed due to missing HDR ACKnowledge from DST node

            return;
            
          case SEND_NO_DACK:
       // Send failed due to missing DATA ACKnowledge from DST node

      return;

        } // switch...

    } // switch...
      
  } else if ( netSendPending ) {
    // Only netSendPending

    gapFound = waitGap(minSendGap);
    if ( gapFound == false ) return;
    
    packetSendCode=sendPacket(nextNetSend);
    switch (packetSendCode) {

      case SEND_OK:
        // Packet successfully sent/acknowledged

        netPacketsSent++;
        sendBuffersFree++;
        sendBufferFree=true;
        packetSent=true;
        break;

      case SEND_CONTENTION:
        // Send failed due to contention detected during SCOUT, quit for now (retrying on next cycle)

        return;

      case SEND_NO_HACK:
        // Send failed due to missing HDR ACKnowledge from DST node

        return;

      case SEND_NO_DACK:
        // Send failed due to missing DATA ACKnowledge from DST node

        return;

    } // switch...
  
  } else {
  // Only netRecvPending

    gapFound = waitGap(minRecvGap);
    if ( gapFound == false ) return; // no minimum GAP found - just quit for now (retrying on next cycle)
      
    packetRecvCode=listenForPacket(maxRecvGap - minRecvGap, lastNetRcvd);
    // NB lastNetRcvd is passed by &reference and points to the MQ slot that matched the received packet

    switch (packetRecvCode) {

      case RECV_OK:
        // Packet received OK (no chksum errors) and (started) within the time-out

        netPacketsRcvd++;
        packetReceived=true;
        break;

      case RECV_FAIL:
        // Packet detected but either chksum mismatch or some other framing error

        return;

      case RECV_UNEXPECTED:
        // Packet received successfully, but not one that we were expecting (Dst/Src & Blk-L/Blk-H)
        
        return;
      
      case RECV_TIMEOUT:
        // No packet received - just quit for now (retrying on next cycle)

        return;

    } // switch...
  
  } // if...
  
  // Now prepare any REPLY, based on packetSent or packetReceived.
  // - in truth, we shouldn't get here unless at least one is true
  
  if ( packetSent ) {
  
    getMQ( nextNetSend, mID, mDev, mAct, mParErr, mLenPar2, mChkSum ); // fetch the message for nextNetSend
    mID = mID | MQ_REPLY_MASK; // Set top bit to mark as REPLY
    mID = mID & ~MQ_LONG_MASK; // Clear bit-14 to mark as SHORT
    mParErr=0;
    mLenPar2=0;
    
    newChkSum = calculateMQChkSum( mID, mDev, mAct, mParErr, mLenPar2 );
    
    putMQ( nextNetSend, mID, mDev, mAct, mParErr, mLenPar2, newChkSum );
    
    nextReply=nextNetSend;
    replyPending = true;
    // digitalWrite(DEBUG_INT5, HIGH); // Switch ON DEBUG_INT5
    
    netSendPending=false;
    // digitalWrite(DEBUG_INT5, LOW); // Switch OFF DEBUG_INT5
    nextNetSend=-1;
    
  } else if ( packetReceived ) {
  
    getMQ( lastNetRcvd, mID, mDev, mAct, mParErr, mLenPar2, mChkSum ); // fetch the message for lastNetRcvd
    mID = mID | MQ_REPLY_MASK; // Set top bit to mark as REPLY
    mID = mID | MQ_LONG_MASK;  // Set bit-14 to mark as LONG
    mParErr=0;
    
    // Discover length of the received packet from the header stored in the first 8 bytes of recvBuffer
    
    mLenPar2 = byte( recvBuffer[ND_NBYT] );
    if ( byte( recvBuffer[ND_TYPE] ) == NF_SBLK ) mLenPar2 = mLenPar2 * 4; // a large FSERVE block 4*ND_NBYT
    
    mLenPar2 += PACKET_HDR_SZ; // don't forget the 8-byte packet header...
    
    netBytesRcvd += mLenPar2;
    
    newChkSum = calculateMQChkSum( mID, mDev, mAct, mParErr, mLenPar2 );
    
    putMQ( lastNetRcvd, mID, mDev, mAct, mParErr, mLenPar2, newChkSum );
    
    nextReply=lastNetRcvd;
    netRecvPendCnt--;
    if ( netRecvPendCnt == 0 ) netRecvPending = false;
    replyPending = true;
    // digitalWrite(DEBUG_INT5, HIGH); // Switch ON DEBUG_INT5
  
  } // if...
  
  return; // all good - a Send or Recv was succesful and the respective message tagged as REPLY
  
} // actionNET()...

boolean waitGap(unsigned int gapTime) {
  // Wait for gapTime us, or return false if any net activity is detected

  timer=0; // Reset the us timer
  while ( timer < gapTime ) {
    
    if ( digitalRead(NET_IN) == HIGH ) return false; // Activity detected on the wire
  
  } // while...
  
  return true; // gapTime reached without any activity
  
} // waitGap...

boolean waitACTive(unsigned int timeOut, unsigned int minDuration) {
  // Wait for NET_IN going ACTive (or already ACTive) - cf. waitRisingEdge
  // - return true if ACTive found within timeOut us, of minDuration length

  unsigned int virtualTimer;
  boolean foundACTive=false;

  virtualTimer=0;
  
  while ( (virtualTimer < timeOut) && (foundACTive == false) ) {
  
    if ( PIND & 0b00000001 ) {
    
      foundACTive=true;
      PORTD |= 0b00000010;
      
    } else {
    
      PORTD &= 0b11111101;
    
    } // if...
    
    virtualTimer++;

  } // while...
  
  if ( ! foundACTive  ) return false; // timeOut reached without detecting ACTive - error
  
  virtualTimer=0;
  
  while ( (virtualTimer < minDuration) && (foundACTive == true) ) {
  
    if ( !(PIND & 0b00000001) ) {
    
      foundACTive=false;
      PORTD &= 0b11111101;
      
    } // if...
    
    virtualTimer++;
  
  } // while...
  
  if ( ! foundACTive ) return false; // timeOut reached without detecting ACTive - error
  
  return true; // Detected ACTive, within timeOut of at least minDuration

} // waitACTive...

boolean waitINACTive(unsigned int timeOut, unsigned int minDuration) {
  // Wait for NET_IN going INACTive (or already INACTive) - cf. waitFallingEdge
  // - return true if INACTive of minDuration length commenced within timeOut us

  unsigned int virtualTimer;
  boolean foundINACTive=false;

  virtualTimer=0;
  
  while ( (virtualTimer < timeOut) && (foundINACTive == false) ) {
  
    if ( PIND & 0b00000001 ) {
    
       PORTD |= 0b00000010;
      
    } else {

      foundINACTive=true;
      PORTD &= 0b11111101;
    
    } // if...
    
    virtualTimer++;

  } // while...
  
  if ( ! foundINACTive  ) return false; // timeOut reached without detecting INACTive - error
  
  virtualTimer=0;
  
  while ( (virtualTimer < minDuration) && (foundINACTive == true) ) {
  
    if ( (PIND & 0b00000001) ) {
    
      foundINACTive=false;
      PORTD &= 0b11111101;
      
    } // if...
    
    virtualTimer++;
  
  } // while...
  
  if ( ! foundINACTive ) return false; // timeOut reached without detecting INACTive - error
  
  return true; // Detected INACTive, within timeOut of at least minDuration

} // waitINACTive...

boolean checkACTive(unsigned int minDuration) {
  // Check for NET_IN remaining ACTive for atleast minDuration

  unsigned int virtualTimer;
  boolean foundINACTive=false;

  PORTD |= 0b00000010;

  virtualTimer=0;
  
  while ( (virtualTimer < minDuration) && (foundINACTive == false) ) {
  
    if ( !(PIND & 0b00000001) ) {
    
      foundINACTive=true;
      
    } // if...
    
    virtualTimer++;
  
  } // while...

  PORTD &= 0b11111101;
  
  if ( foundINACTive ) return false; // minDuration of ACTive not reached error
  
  return true; // Detected continuous ACTive of minDuration

} // checkACTive...

boolean waitRisingEdge(unsigned int timeOut) {
  // Wait for NET_IN going INACTive to ACTive (NOT already ACTive) - cf. waitACTive
  // - return true (ACTive edge found within tinmeOut us) or false

  boolean foundINACTive=false;
  
  timer=0; // Reset the us timer    
  while ( timer < timeOut ) {
  
    if ( digitalRead(NET_IN) == LOW ) foundINACTive=true;
  
  } // while...
  
  if ( ! foundINACTive ) return false; // INACTive not found in timeOut us
  
  while ( timer < timeOut ) {
  
    if ( digitalRead(NET_IN) == HIGH ) return true; // Rising edge  detected on the wire
  
  } // while...
  
  return false; // timeOut reached Rising edge detected - error

} // waitRisingEdge...

boolean waitFallingEdge(unsigned int timeOut) {
  // Wait for NET_IN going ACTive to INACTive 
  // - return true (INACTive edge found within timeOut us) or false

  unsigned int virtualTimer;
  boolean foundACTive=false;
  boolean foundFallingEdge=false;

  virtualTimer=0;
    
  while ( (virtualTimer < timeOut) && (! foundACTive) ) {
  
    if ( PIND & 0b00000001 ) {
    
      foundACTive=true;
      PORTD |= 0b00000010;
      
    } else {
    
      PORTD &= 0b11111101;
      
    } // if...
    
    virtualTimer++;
      
  } // while...
  
  if ( ! foundACTive ) return false; // ACTive not found in timeOut us
  
  while ( (virtualTimer < timeOut) && (foundFallingEdge == false) ) {
  
    if ( !(PIND & 0b00000001) ) {
    
      PORTD &= 0b11111101;
      foundFallingEdge=true; // Falling edge detected on the wire
      
    } else {
    
      PORTD |= 0b00000010;
      
    } // if...
    
    virtualTimer++;
  
  } // while...
  
  if ( foundFallingEdge ) return true;
  
  return false; // timeOut reached without Falling edge detected - error

} // waitFallingEdge...

byte listenForPacket(unsigned int timeOut, int &recvMQSlot) {
  /* Listen for a packet on the wire for timeOut us
   * - depending upon the outcome, return the byte value:
   *   RECV_OK=0         - Packet received and chksums matched - 
   *   RECV_FAIL=1       - Packet received but either the header/data chksums didn't match, or some framing error
   *   RECV_UNEXPECTED=2 - Packet received OK, but did not match any pending receive maessages
   *   RECV_TIMEOUT=3    - No packet detected in timeOut us
   * - For a successful receive, update the recvMQSlot (ref) parameter with the MQ Slot ID that matched the packet header.
  */

  boolean foundSCOUT, canSendACK;
  
  byte rcvdDst, rcvdSrc, rcvdBlkL, rcvdBlkH, rcvdType, rcvdNByt, rcvdDChk, rcvdHChk;
  byte recvCode, calcHChk, calcDChk;
  
  byte hdrCheck;
  unsigned int packetLen;
  
  if ( netRecvPendCnt < 1 ) return RECV_UNEXPECTED; // Should NOT happen...
  
  noInterrupts();
  
  foundSCOUT=waitACTive(timeOut, SCOUT_DETECT_TIME);
  if ( ! foundSCOUT ) {
  
    interrupts();
    return RECV_TIMEOUT; // SCOUT not detected, or wasn't long enough
    
  } // if...

  delayMicroseconds(SCOUT_SKIP_TIME); // Skip-over the SCOUT (for now - I've got ideas...)
  
  recvCode = recvBytes(0, PACKET_HDR_SZ, calcHChk); // NB calcHChk sums one too many bytes - correct in checkHeader.
  
  if ( recvCode != RECV_OK ) {
  
    interrupts();
    return RECV_FAIL;
    
  } // if...
  
  // Capture and check Hdr ChkSum
  hdrCheck = checkHeader( recvMQSlot, rcvdDst, rcvdSrc, rcvdBlkL, rcvdBlkH, rcvdType, rcvdNByt, rcvdDChk, rcvdHChk, calcHChk);
  
  switch (hdrCheck) {
  
    case HDR_GOOD:
      // The Packet Hdr was as anticipated and received without error (ChkSums matched)
      // - proceed to ACKnowlegde and read-in the actual packet payload.
    
      //delayMicroseconds(PRE_ACK_DELAY);
      sendByte(VALID_ACK, false); // Send ACK and don't leave HIGH
      break;
    
    case HDR_CHKSUM_ERROR:
     // A packet arrived, but the ChkSum we calculated didn't match what we found in the Header
     // - we can't say for sure whether the packet was really destined for us or desired.
     //   Ignore and return.
            
PORTB |= 0b00000001;
delayMicroseconds(50);
PORTB &= 0b11111110;

      interrupts();
      return RECV_FAIL;
      
    case HDR_NOT_FOR_ME:
      // The Packet Hdr was received but was either not for us (DST <> SELF)
      // - or else we weren't expecting a packet from SRC...
      //   Ignore and return.
    
      interrupts();
      return RECV_UNEXPECTED;
    
    case HDR_PHANTOM_BLOCK:
      // A ghost from the past was received - a packet that we had previously wanted, received and ACKnowledged
      // - presumably, the sender failed to recognise our earlier ACK and has resent the (n-1) packet
      //   We'll ACK again patiently, but otherwise take no further action - quietly awaiting the next expected packet
      //  Note to self: we might want to leave the very last RECV message request for a given session 'open'
      //    for a time after ACK'ing, to accomodate this very situation and release the sender sooner than the standard QDOS protocol would...
    
      //delayMicroseconds(PRE_ACK_DELAY);
      sendByte(VALID_ACK, false); // Send ACK and don't leave HIGH
      interrupts();
      return RECV_TIMEOUT;
      
    default:
      // Shouldn't get to this, but...   
      
      interrupts();
      return RECV_TIMEOUT;
    
  } // switch...
  
  // If we arrive here at all, we're ready to accept a packet meant for me.
  // - recvMQSlot will have been populated by checkHeader and point to the MQ appropriate for the REPLY
  
  if ( rcvdType == NF_SBLK ) {
  
    packetLen = int(rcvdNByt) * 4; // This appears to be an FSERVE large Block - x4 (4 - 1020 bytes)
    
  } else {
  
    packetLen = int(rcvdNByt);
  
  } // if...
  
  if ( packetLen > MAX_PACKET_SZ ) {
   
    interrupts();
    return RECV_FAIL; // This isn't actually a recv error, but an unexpectedly long packet...
    
  } // if...
  
  recvCode = recvBytes(PACKET_HDR_SZ, packetLen, calcDChk); // Read in to recvBuffer from PACKET_HDR_SZ onwards
  
  if ( recvCode !=RECV_OK ) {
    // Don't ACK and simply return a read failure
      
PORTB |= 0b00000001;
delayMicroseconds(5);
PORTB &= 0b11111110;

    interrupts();
    return RECV_FAIL;
    
  }
  
  // OK, let's calculate the DATA chksum
  
  //calcDChk=0;
  //for ( unsigned int index=0; index < packetLen; index++ ) calcDChk += recvBuffer[PACKET_HDR_SZ + index];
  
  if ( calcDChk != rcvdDChk ) {
    // DATA ChkSum didn't match expected, return read failure...
      
PORTB |= 0b00000001;
delayMicroseconds(15);
PORTB &= 0b11111110;

    interrupts();
    return RECV_FAIL;
    
  }
  
  // OK... Ao let's wait for end of the last STOP and then send the ACK, returning as read success.

  // Added v1.8h (following testing with ZXNet) - use waitINACTive in place of simple delay - before sending ACK
  // With the added check for a minimum gap
  // - setting reasonable defaults here - wait for a full bytes-worth of silence, starting within 3x BIT-times

  canSendACK = waitINACTive( BIT_TIME_RECV * 3 , BIT_TIME_RECV * 10 );

  if ( ! canSendACK ) {

    interrupts();
    return RECV_FAIL;
    
  }
  
  sendByte(VALID_ACK, false); // Send ACK and don't leave HIGH
  
  recvBuffersFree--;
  if ( recvBuffersFree < 1 ) recvBufferFree=false;
  
  // The REPLY itself (referenced by recvMQSlot) will be updated by the calling actionNet routine
  //   based on the packet header at recvBuffer
  
  interrupts(); // And the all-important re-enabling of interrupts...
  return RECV_OK;

} // listenForPacket...

byte checkHeader( int &matchingMQSlot, byte &dst, byte &src, byte &blkL, byte &blkH, byte &typ, byte &nbt, byte &dck, byte &hck, byte calcHChk) {
  // Capture first 8 bytes of recvBuffer in to (ref) params and check the Header ChkSum
  // - proceed to validate the DST, SRC, BLKL, BLKH against all the RECVieve-type MQ entries
  
  boolean foundRecvEntry=false;
  byte mDev, mAct;
  unsigned int mID, mParErr, mLenPar2, mChkSum;
  
  dst  = byte( recvBuffer[ND_DEST] );
  src  = byte( recvBuffer[ND_SELF] );
  blkL = byte( recvBuffer[ND_BLKL] );
  blkH = byte( recvBuffer[ND_BLKH] );
  typ  = byte( recvBuffer[ND_TYPE] );
  nbt  = byte( recvBuffer[ND_NBYT] );
  dck  = byte( recvBuffer[ND_DCHK] );
  hck  = byte( recvBuffer[ND_HCHK] );
  
  //hckCalculated=0;
  //for ( byte index=0; index < (PACKET_HDR_SZ - 1); index++ ) hckCalculated += recvBuffer[index];
  
  calcHChk -= hck;
  
  if ( hck != calcHChk ) {
  
    return HDR_CHKSUM_ERROR;
    
  } // if...

  for ( int mQSlot=0; mQSlot <= MAX_MESSAGES; mQSlot++ ) {
  
    getMQ( mQSlot, mID, mDev, mAct, mParErr, mLenPar2, mChkSum ); // fetch the message for mQSlot

    if ( !( mDev == M_DEV_NET && ( ( mAct == M_ACT_RECEIVE ) || ( mAct == M_ACT_SERV_START ) ) ) ) {
      
PORTB |= 0b00000001;
delayMicroseconds(5);
PORTB &= 0b11111110;

      continue;
    
    } // if...
    
    if ( word(dst, src) != mParErr ) {
    
PORTB |= 0b00000001;
delayMicroseconds(15);
PORTB &= 0b11111110;

      continue; // Wasn't for me, or from whom I expected it...
      
    } // if...
    
    foundRecvEntry=true;
    matchingMQSlot=mQSlot;
    break;

    // The structure of this for-loop is a bit messy due to the continue's effectively re-routing flow
    // - should probably be a do...while instead, but who's watching?
    
  } // for...
  
  if ( !foundRecvEntry ) {
    
PORTB |= 0b00000001;
delayMicroseconds(25);
PORTB &= 0b11111110;

    return HDR_NOT_FOR_ME;
    
  } // if...
  
  // Now, for ease of numeric comparison of BLK num (below is the ONLY such requirement),
  //   at both ends of the MQ (USB Host and here), we arrange for mLenPar2 to be in the big-endian form (BlkH, BlkL)
  //   rather than how they appear over the wire (which was originally to suit ZXNet running on the Z80)
  // Note to self PLEASE remember this when coding the MQ in SMSQ...
  
  if ( word(blkH, blkL) == (mLenPar2 - 1) ) {
    
PORTB |= 0b00000001;
delayMicroseconds(35);
PORTB &= 0b11111110;
  
    return HDR_PHANTOM_BLOCK;
    
  } // if...
    
  if ( word(blkH, blkL) != (mLenPar2) ) {
    
PORTB |= 0b00000001;
delayMicroseconds(45);
PORTB &= 0b11111110;
  
    return HDR_NOT_FOR_ME;
    
  } // if...
    
  /// Phew! Got there in the end. Update the ref param matchingMQSlot and return read success
    
PORTB |= 0b00000001;
delayMicroseconds(55);
PORTB &= 0b11111110;

  return HDR_GOOD;
  
} // checkHeader...

byte recvBytes(int bufferOffset, unsigned int byteCnt, byte &chkSum) {
  /* Listen for byteCnt bytes and populate recvBuffer - the bufferOffset allows us to post the data as HDR, then DATA
   * - 2xLEAD + 1xSTART + 8xBITs + 2xSTOP (+1x FILL between consecutive bytes)
   *   RETurns an error code - which we don't use much today, and a ref parm chkSum
  */
  
  byte nextByte, recvCode;
  unsigned int leadWaitTime;
  
  leadWaitTime=LEAD_WAIT_TIME;
  
  chkSum = 0;
  
  for ( unsigned int bytePtr=0; bytePtr < byteCnt; bytePtr++ ) {

    recvCode = recvByte(leadWaitTime, nextByte);
    
    switch (recvCode) {
    
      case RECV_OK:
        
        recvBuffer[bufferOffset + bytePtr] = nextByte;
        chkSum += nextByte;
        break;
        
      case RECV_FAIL:
        // Framing error (START or STOP not detected) - just fall-through to RECV_FAIL...
      
      case RECV_TIMEOUT:
        // Timeout waiting for LEAD bits (could be less bytes in packet than there should be...)
        
        return recvCode; // Pass-back immediately to listenForPacket
      
    } // switch...

    leadWaitTime=LEAD_WAIT_TIME_2;
    
  } // for...
   
  return RECV_OK; // Wow! We successfully captured a whole block (HDR or DATA)!
  
} // recvBytes...

byte recvByte(unsigned int leadWaitTime, byte &receivedByte) {
  /* Wait for a ACTive LEAD bits in leadWaitTime (LEAD_WAIT_TIME for first byte/ACK, then nominal) us,
   * followed by INACTive START bit in START_WAIT_TIME
   * - 8x bits (LSBit first), followed by atleast 2 STOP bits (make cursory check, MIN does, but Tk2 doesn't)
   * - receivedByte returned by reference, error returned directly
  */
  
  boolean foundLEAD, foundSTART, foundSTOP;  
  // byte bitSampleCnt=0;
  // unsigned int virtualTimer;

  foundLEAD = waitACTive(leadWaitTime, BIT_TIME_RECV - 5); // Check for at least 1x BIT_TIME of LEAD (should be atleast 2 in practice)
  if ( ! foundLEAD ) return RECV_TIMEOUT;

  foundSTART = waitFallingEdge(START_WAIT_TIME);
  if ( ! foundSTART ) return RECV_FAIL;

  delayMicroseconds(BIT0_TIME_RECV); // pause to centre of bit-0 (taking account of delay in detection of START bit)
  
  receivedByte=0; // reset the reference parameter
  
  // virtualTimer=0; // reset our virtual timer, ready to capture byte from NET_IN
    
  for ( byte b=1; b < 9; b++ ) {
  
    // for ( byte sample=0; sample < 5; sample++ ) {
    //  if ( digitalRead(NET_IN) == HIGH ) bitSampleCnt++;
    //}
    
    PORTB |= 0b00000001; // turn ON DEBUG_INT5
    
    receivedByte = receivedByte >> 1; // shifted BEFORE adding next bit (avoids complications with 8x shifts)
    
    // if ( bitSampleCnt >= 3 ) receivedByte = receivedByte | 0b10000000; // NB Bit-7 set
    // if ( bitSampleCnt >= 1 ) {
    
    if ( PIND & 0b00000001 ) {
    
      receivedByte = receivedByte | 0b10000000; // NB Bit-7 set
      PORTD |= 0b00000010; // Switch ON DEBUG_NET (fast)
      
    } else {
    
      PORTD &= 0b11111101; // Switch OFF DEBUG_NET (fast)
    
    }
    
    // while ( timer < (b * BIT_TIME) ); // spin until next BIT_TIME expires
    
    delayMicroseconds(BIT_TIME_RECV);
    
    PORTB &= 0b11111110;  // turn OFF DEBUG_INT5
    
    //bitSampleCnt=0; // reset for next bit...
  
  } // for...
  
  // We are now at the centre of first STOP bit
    
  foundSTOP = checkACTive(BIT_TIME_RECV - 4); // Check for at least 1x BIT_TIME of STOP
  if ( ! foundSTOP ) return RECV_FAIL;
  
  return RECV_OK; // Ref. parameter receivedByte has been created from bit samples

} // recvByte...

byte sendPacket(int sendMQSlot) {
  /* Attempt to transmit the packet referenced by sendMQSlot, with payload in sendBuffer
   * - depending upon the outcome, return the byte value:
   *   SEND_OK=0         - Packet sent and acknowledged by DST
   *   SEND_CONTENTION=1 - Contention detected during SCOUT phase
   *   SEND_NO_HACK=2    - Header sent, but was not acknowldged by DST
   *   SEND_NO_DACK=3    - Data packet sent, but was not acknowldged by DST
   * - GAP has already been found, so start with SCOUT (checking for contention) immediately...
  */

  boolean scoutCode;
  byte mDev, mAct, recvCode, acknowledgeCode;
  unsigned int mID, mParErr, mLenPar2, mChkSum, packetLen;
  
  getMQ( sendMQSlot, mID, mDev, mAct, mParErr, mLenPar2, mChkSum ); // fetch the message for sendMQSlot

  scoutCode = sendSCOUT(SCOUT_WORD);
  
  if ( ! scoutCode ) return SEND_CONTENTION; // Detected contention during SCOUT
  
  // Might want to extend SCOUT ACTive a little here - let's start without...
  
  noInterrupts();
  
  sendBytes(0, PACKET_HDR_SZ); // Transmit packet Header (8x bytes from index 0 of sendBuffer)

  // Comment-out next if clause when testing - disables waiting for a HACK... 
  recvCode = recvByte(LEAD_WAIT_TIME, acknowledgeCode); // Presently only expecting 0x01 in response
  // if ( recvCode != RECV_OK ) return SEND_NO_HACK;
  if ( recvCode != RECV_OK || acknowledgeCode != VALID_ACK ) {
  
    interrupts();
    return SEND_NO_HACK;
    
  } // if...
  
  // OK, got a positive/valid ACK, let's pause a little before sending packet
  delayMicroseconds(POST_ACK_DELAY);
  
  packetLen = mLenPar2 - PACKET_HDR_SZ;
  sendBytes(PACKET_HDR_SZ, packetLen); // Transmit packet Data (packetLen bytes from index 8 of sendBuffer)

  // Comment-out next if clause when testing - disables waiting for a DACK...
  recvCode = recvByte(LEAD_WAIT_TIME, acknowledgeCode); // Presently only expecting 0x01 in response
  //if ( recvCode != RECV_OK ) return SEND_NO_DACK;
  
  interrupts();
      
  if ( recvCode != RECV_OK || acknowledgeCode != VALID_ACK ) { 

    return SEND_NO_DACK;

  } else {
    // Phew! Seems to have sent successfully - return OK

    netBytesSent += PACKET_HDR_SZ + packetLen; // only update netBytesSent if packet was fully ACKnowledged.
    return SEND_OK;
    
  } // if...

} // sendPacket...

boolean sendSCOUT(unsigned int scoutWord) {
  /* Transmit SCOUT based on the scoutWord, checking for contention
   * - Uses SCOUT_BIT_TIME inplace of BIT_TIME
   *   scoutWord could be used in future implementations to flag capabilities for compatible clients
   *   Returns true if transmitted succesfully with no contention detected (only detectable durng NET_INACTive)
  */

  boolean scoutBitState;
  unsigned long scoutExtTime;
  
  scoutWord = scoutWord | 0b1000000000000011; // Ensure 2x START and 1x STOP
  
  timer=0; // Reset the us timer
  
  for ( byte scoutBit=0; scoutBit < SCOUT_BITS; scoutBit++ ) {

    scoutBitState = ! ( scoutWord & 0b0000000000000001 ); // NB Inverted logic suitable for PNP transitor    
    digitalWrite(NET_OUT, scoutBitState);
    
    if ( ! scoutBitState ) {
    
      // Can't check for contention during NET_ACTive
      delayMicroseconds(SCOUT_BIT_TIME);
      
    } else {
    
      // Check for contention whilst NET_INACTive, 5x times

      for ( byte chkContention=0; chkContention < 5; chkContention++ ) {

        delayMicroseconds(SCOUT_BIT_TIME_DIV5);      
        if ( digitalRead(NET_IN) == HIGH ) return false; // Contention detected, return false
        
      } // for...
      
    } // if...
    
    scoutWord = scoutWord >> 1;
  
  } // for...

  // Theoretically, 16x 30uS (SCOUT_BIT_TIME) = 480us have passed - we need to leave it ACTive until atleast 525us
  
  scoutExtTime = (SCOUT_DETECT_TIME + SCOUT_SKIP_TIME) - timer;
  delayMicroseconds(scoutExtTime);
  
  return true; // SCOUT transmitted without contention being detected, return OK (NET_OUT still ACTive)

} // sendSCOUT...

void sendByte(byte nextByte, boolean leaveActive) {
  /* Transmit a single byte, leaving the NET_OUT port at the value of leaveActive
   * - 2xLEAD + 1xSTART + 8xBITs + 2xSTOP ( + 1x FILL if leaveActive )
   * - NB - NET_OUT needs the logical inverse using NET_ACTIve (false/LOW) and NET_INACTive (true/HIGH)
   *        due to the PNP transitor hardware used in the output circuit (matching QLAN design).
  */
  boolean bitState;
  
  timer=0;
  
  digitalWrite(NET_OUT, NET_ACTive);
  delayMicroseconds(LEAD_BITS * BIT_TIME_SEND);
  
  digitalWrite(NET_OUT, NET_INACTive);
  delayMicroseconds(START_BITS * BIT_TIME_SEND);

  for ( byte bit=0; bit < 8; bit++ ) {

    bitState = !(nextByte & 0b00000001); // NB Inverted logic suitable for PNP transitor
    digitalWrite(NET_OUT, bitState);
    delayMicroseconds(BIT_TIME_SEND);
    nextByte = nextByte >> 1;
    
  } // for...
  
  digitalWrite(NET_OUT, NET_ACTive);
  delayMicroseconds(STOP_BITS * BIT_TIME_SEND);
  
  if ( leaveActive ) {

    delayMicroseconds(FILL_BITS * BIT_TIME_SEND - 4);
    
  } else {
  
    digitalWrite(NET_OUT, NET_INACTive);
  
  } // if...
  
} //sendByte...

void sendBytes(int bufferOffset, int byteCnt) {
  /* Transmit byteCnt bytes from sendBuffer - the bufferOffset allows us to pick the data from the header
   * - 2xLEAD + 1xSTART + 8xBITs + 2xSTOP (+1x FILL between consecutive bytes)
  */
  
  boolean leaveActive=true;
  byte nextByte;
  
  for ( int bytePtr=0; bytePtr < byteCnt; bytePtr++ ) {

    nextByte=sendBuffer[bufferOffset + bytePtr];
    if ( bytePtr == (byteCnt - 1) ) leaveActive=false;

    sendByte(nextByte, leaveActive);
    // Will add a FILL_BITS delay with NET_OUT left ACTive after each byte, except the very last
    
  } // for...
  
} // sendBytes..

void replyMessage() {
  /* Send over USB the next REPLY from the MQ based on nextReply
   *  
  */

  byte mDev, mAct;
  unsigned int mID, mParErr, mLenPar2, mChkSum;

  mQSlotAvailable=false;

  if ( nextReply == -1 ) {
  
    replyPending=false;
    // digitalWrite(DEBUG_INT5, LOW); // Switch OFF DEBUG_INT5
    return; // Make quick return if, for some reason, the nextReply is not set (-1)
    
  } // if...

  // nextReply should therefore hold the mQSlotIndex in to the Message Queue, ready for REPLY

  getMQ( nextReply, mID, mDev, mAct, mParErr, mLenPar2, mChkSum ); // Get some details about the REPLY

  // Serial.flush(); // On Teensy, will wait until all previous but still pending Sends are completed, or return immediately
  
  flashLED_start(LED_PIN_USB_TX);  
  Serial.write( MQ_SYNC, MQ_SYNC_LEN );
  Serial.write( &messageQueue[nextReply * MESSAGE_HLEN], MESSAGE_HLEN );

  if ( mLenPar2 > 0 ) {
    Serial.write( recvBuffer, mLenPar2 );
    recvBuffersFree++;
    recvBufferFree=true;
  }

  Serial.send_now(); // On Teensy, will force (and wait for) all pending USB data to be transmitted

  usbBytesSent+= MQ_SYNC_LEN + MESSAGE_HLEN + mLenPar2;

  markMQSlotEmpty( nextReply ); // Flag this MQ Slot entry to show processed/now empty

  // Find next free MQ Slot (dupe of what we do in queueMessage)
  for ( int mQSlot=0; mQSlot <= MAX_MESSAGES; mQSlot++ ) {

    getMQ( mQSlot, mID, mDev, mAct, mParErr, mLenPar2, mChkSum );
    
    if ( mID == 0 ) {
    
      nextFreeMQSlot=mQSlot; // Find the lowest free MQ Slot
      mQSlotAvailable=true;
      break; // This slot is 'empty' - we're done, leave for loop
      
    } // if...

  } // for...
  
  nextReply = -1;
  replyPending = false;
  // digitalWrite(DEBUG_INT5, LOW); // Switch OFF DEBUG_INT5
  
} // replyMessage()...

void copyArray( byte *srcArr, byte *dstArr, int srcIdx, int dstIdx, int byteCnt ) {

  int byteIdx;

  for ( byteIdx=0; byteIdx < byteCnt; byteIdx++ ) {
    dstArr[dstIdx] = srcArr[srcIdx];
    dstIdx++;
    srcIdx++;
  }
  
} // copyArray()...

void loop() {
  // put your main code here, to run repeatedly:

  initialiseVariables();

  while (!Serial.dtr()) ; // Wait for program running on host PC to open COM port

  flashLEDs(3, 500);

  do {
    
    digitalWrite(DEBUG_INT1, HIGH); // switch ON the DEBUG_INT1 port for testing
    
    queueMessage();
    
    digitalWrite(DEBUG_INT1, LOW); // switch OFF the DEBUG_INT1 port for testing
    
    if ( mQScanRequired ) {
    
      // digitalWrite(DEBUG_INT2, HIGH);
      scanMessageQueue();
      // digitalWrite(DEBUG_INT2, LOW);
      
    } // if...

    if ( netActionPending ) {
    
      digitalWrite(DEBUG_INT3, HIGH);
      actionNET();
      digitalWrite(DEBUG_INT3, LOW);
      
    } // if...

    if ( replyPending ) {

      digitalWrite(DEBUG_INT4, HIGH); // switch ON the DEBUG_INT4 port for testing    
      replyMessage();
      digitalWrite(DEBUG_INT4, LOW); // switch OFF the DEBUG_INT4 port for testing

    } // if...
    
    digitalWrite(DEBUG_INT2, HIGH);
    flashLED_async();
    digitalWrite(DEBUG_INT2, LOW);
    
  } while (cmd != M_ACT_RESET);
  
} // loop()...
